// modbus_slave_example.c
// c. 2018 Technologic Systems
// written by Michael Peters

// This example sets up a modbus slave node that listens for modbus commands.

// The modbus_map structure is documented literally nowhere.  I dug it out of the source
//  code in modbus.h:
/*   THIS STRUCT IS HERE FOR REFERENCE.  IT IS DEFINED IN modbus.h.
typedef struct {
    int nb_bits;
    int nb_input_bits;
    int nb_input_registers;
    int nb_registers;
    uint8_t *tab_bits;
    uint8_t *tab_input_bits;
    uint16_t *tab_input_registers;
    uint16_t *tab_registers;
} modbus_mapping_t;
*/

// Compile instructions:
//  Install package pkg-config.
//  gcc `pkg-config --cflags --libs libmodbus` <filename.c> -o <binaryname> -Wall

#include <stdio.h>
#include <stdlib.h>
#include <modbus.h>
#include <errno.h>
#include "modbus_example.h"

#define SERIAL_PORT "/dev/ttts3"
#define CONFIG_FILE "slaveid.conf"

modbus_t* comm_setup();  // sets up modbus port.
int get_slave_id_from_file(); // Grabs slave ID from config file.

int main(int argc, char **argv)
{
	int err;
	// set up the modbus data registers structure.
	modbus_mapping_t *modbus_map;
	modbus_map = modbus_mapping_new(0, 0, 0x14, 0);  // all values init to zero.
	if(modbus_map == NULL)
		return 1;  // Stop.  NULL modbus_map means ENOMEM.
	modbus_map->tab_registers[DA_REGISTER] = 0x00DA; // This is a test register.  Always returns 0xDA.

	modbus_t *mb = comm_setup(); // turn on the serial port & set up *mb.
	if(!mb){
		modbus_mapping_free(modbus_map);
	 	return 1;
	}

	int slave_id = get_slave_id_from_file();  // Grab slave ID from file.
	if(slave_id <= 0){
		fprintf(stderr, "slave_id = %d\n", slave_id);
		fprintf(stderr, "Error opening slave ID config file.\n");
		modbus_mapping_free(modbus_map);
		return 1;  // quit if error opening slave ID file.	
	} 
	else modbus_set_slave(mb, slave_id);

	modbus_set_debug(mb,TRUE);  // This causes libmodbus to output all packet data to stdout.

	uint8_t req[180]; // The actual packet.  Should be larger than any inbound packet.
	int num_reqs;
	int sequence_number;
	while(TRUE){
		// listen for modbus requests.
		num_reqs = modbus_receive(mb, req);  // default blocking request will not time out.
		if(num_reqs == -1){
			fprintf(stderr, "modbus_receive() returned -1 status.\nerrno=%d\n%s\n", 
				errno, modbus_strerror(errno));
		}

		// Respond to modbus request.  This must be run immediately after modbus_receive.
		err = modbus_reply(mb, req, num_reqs, modbus_map);
		if(err == -1) printf("Error sending reply.  %s\n", modbus_strerror(errno));
		else printf("Sent reply.  Length %d.\n", err); // if err is not -1 then err=length.

		//  Any new data should now be in the modbus map for the slave to process.

		// Check for messages in the message register.
		if(modbus_map->tab_registers[MESSAGE] != 0){
			printf("Message from Modbus Master is 0x%04X.\n", modbus_map->tab_registers[MESSAGE]);
			if(modbus_map->tab_registers[MESSAGE] == REBOOT_ORDER)
				printf("Message contained reboot command.  Replace this printf with a safe shutdown and reboot.\n");
			modbus_map->tab_registers[MESSAGE] = 0; // Reset the messages register after acting on it.
		}
		else
			printf("Nothing in the message register.\n");

		// check to see if there is a new sequence number being requested.
		if(modbus_map->tab_registers[NEW_SEQUENCE] != 0){
			sequence_number = modbus_map->tab_registers[NEW_SEQUENCE];
			modbus_map->tab_registers[CURRENT_SEQUENCE] = sequence_number;
			// reset the new_sequence register.
			modbus_map->tab_registers[NEW_SEQUENCE] = 0;
		}

		// This is the end of the packet processing loop.
	}

	modbus_mapping_free(modbus_map);
	modbus_close(mb);
	modbus_free(mb);
	return 0;
}

// Opens CONFIG_FILE and reads in a text value.
//  If there's a problem, returns -1.  Otherwise returns a number.
int get_slave_id_from_file()
{
	int the_id;
	FILE *fp = fopen(CONFIG_FILE, "r");
	int numbytes;

	if(fp == NULL){
		the_id=-1;
	}
	else{
		numbytes = fscanf(fp, "%d", &the_id);
		if(numbytes==0)
			the_id=-1;
	}
	return the_id;
}

// Creates serial port connection and allocates modbus port.
//  Returns a pointer to the modbus context (ctx).
modbus_t* comm_setup()
{
	modbus_t *mb;
	mb = modbus_new_rtu(SERIAL_PORT, BAUD_RATE, 'N', 8, 1);
	if(!mb){
		fprintf(stderr, "unable to open bus at port %s.\n", SERIAL_PORT);
	}
	modbus_connect(mb);
	return mb;
}
